/* Przemysaw Boruta, grupa 4I1
 *
 * Wzorzec projektowy wizytator
 *
 * Plik zawiera przykad uycia wzorca projektowego "wizytator".
 * Wzorzec ten umoliwia definiowanie dziaa na strukturze obiektowej
 * bez modyfikacji klas nalecych do tej struktury. Osigane jest to
 * przez "kapsukowanie" operacji wykonywanych na obiektach w oddzielnej
 * hierarchii klas wizytatorw.
 *
 * Hierarchia klas odwiedzanych reprezentuje uproszczony asortyment sklepu spoywczego.
 * Na jej grze znajduje si abstrakcyjna klasa Produkt, deklarujca czysto wirtualn
 * metod akceptuj, odpowiedzialn za wywoanie algoryutmu wizytacji, zdefiniowanego 
 * w klasie wizytatora. Klasy z niej dziedziczce przedefiniowuj t metod 
 * w odpowiedni dla siebie sposb.
 *
 * Konkretne produkty (chleb, piwo, wielopak) mog by odwiedzane przez obiekty klas
 * znajdujcych sie w hierarchii klas wizytujcych. Na jej grze znajduje si abstrakcyjna
 * klasa Wizytator, deklarujca czysto wirtualne metody 'wizytuj', odpowiedzialne za
 * wywoanie algorytmu wizytacji dla kadej z klas. Klasy konkretnych wizytatorw (tu:
 * wizytatory odpowiedzialne za wypisanie produktu na ekran oraz zliczenie wartoci
 * produktw znajdujcych si w koszyku) definiuj kad z tych metod w odpowiedni 
 * dla siebie sposb.
 * */

#include <iostream>
#include <vector>
using namespace std;

class Chleb;
class Piwo;
class Wielopak;

/* Klasa bazowa dla hierarchii klas wizytatorw. Czysto wirtualne metody 'wizytuj'
 * s definiowane w konkretnych klasach wizytatorw.
 *
 */
class Wizytator{
	public:
		virtual void wizytuj(Chleb&) = 0;
		virtual void wizytuj(Piwo&) = 0;
		virtual void wizytuj(Wielopak&) = 0;
};

/* Klasa bazowa hierarchii klas produktw. Czysto wirtualna metoda 'akceptuj' jest 
 * definiowana w kadej z konkretnych klas produktw. Podanie jako jej argumentu 
 * referencji na klas Wizytator umoliwia przyjcie przez klas reprezentujc
 * produkt dowolnego wizytatora z hierarchii wizytatorw.
 *
 * Klasa uwzgldnia take cen danego produktu.
 */
class Produkt{
	public:
		virtual void akceptuj(Wizytator& w) = 0;
		double podajCene() { 
			return cena;
		}
	protected:
	
		Produkt(double gcena) : cena(gcena) {}
	private:
		double cena;
};

/* Klasa reprezentujca konkretny produkt, w tym wypadku chleb. Jej metoda 'akceptuj' polega
 * na prostym wywoaniu metody 'wizytuj' obiektu wizytujcego. Obiekt typu Chleb nie wie,
 * jakiego typu wizytator zosta przyjty. 
 *
 * Obiekt wizytujcy posiada dostp do obiektu wizytowanego poprzez referencj, przekazan mu
 * w wywoaniu funkcji 'wizytuj'. Dziki temu ma moliwo wykonania zdefiniowanego
 * algorytmu.
 */
class Chleb : public Produkt{
	public:
		Chleb(double gcena, double gwaga) : Produkt(gcena), waga(gwaga) {}
		// Definicja odziedziczonej funkcji czysto wirtualnej
		virtual void akceptuj(Wizytator& w){
			// Wywoanie algorytmu wizytacji, zdefiniowanego w obiekcie wizytatora
			w.wizytuj(*this);
		}
		double podajWage() {
			return waga;
		}

	private:
		double waga;
};

/* Klasa podobna pod wzgldem omawianych zagadnie do klasy Chleb. Wywoanie algorytmu wizytacji rwnie
 * polego na prostym wywoaniu funkcji 'wizytuj' obiektu wizytujcego z argumentem bdcym referencj
 * do obiektu wizytowanego, co umoliwia wizytatorowi wykonanie dziaania.
 */
class Piwo : public Produkt{
	public:
		Piwo(double gcena, double gwielkoscPuszki) : Produkt(gcena), wielkoscPuszki(gwielkoscPuszki) {}
		virtual void akceptuj(Wizytator& w){
			w.wizytuj(*this);
		}
		double podajWielkoscPuszki(){
			return wielkoscPuszki;
		}
	private:
		double wielkoscPuszki;
};

/* Klasa ta demonstruje wywoanie obiektu wizytujcego dla obiektu bdcego zbiorem innych obiektw.
 * W tym przypadku najpierw wywoywany jest algorytm wizytacji dla samej klasy kontenera (np. wypisanie
 * na ekranie informacji o 'zbiorczym' charakterze obiektu) a nastpnie wywoywana jest wizytacja dla 
 * wszystkich obiektw znajdujcych si w danym kontenerze.
 */
class Wielopak : public Produkt{
	public:
		Wielopak(double gcena, int gilosc) : Produkt(0), ilosc(gilosc) {
			for(int i=1; i<= gilosc; i++)
				wek.push_back(Piwo(gcena/gilosc,500));
		}
		virtual void akceptuj(Wizytator& w){
			//Wizytacja samego kontenera
			w.wizytuj(*this);
			for(vector<Piwo>::iterator i = wek.begin(); i != wek.end(); i++)
				//Wizytacja skadowych
				w.wizytuj(*i);
		}
		int podajIlosc() {
			return ilosc;
		}
	private:
		vector<Piwo> wek;
		int ilosc;
};

/* Klasa wizytatora odpowiedzialnego za wypisanie na ekran informacji o obiekcie podlegajcym wizytacji.
 * Definiuje ona wszystkie funkcje zwizene z wizytacj konkretnych typw produktw, wypisujc informacje
 * nieco inaczej jeli wypisywany obiekt jest chlebem, piwem lub wielopakiem.
 */
class Wypisanie : public Wizytator{
	public:
		// Metoda wywoywana podczas wizytacji obiektu klasy Chleb
		virtual void wizytuj(Chleb& c){
			cout << "+ Chleb o wadze " << c.podajWage() << " gramow  w cenie " << c.podajCene() << " zl" <<  endl;
		}
		// Metoda wywoywana podczas wizytacji obiektu klasy Piwo 
		virtual void wizytuj(Piwo& p){
			cout << "+ Piwo w puszce o wielkosci " << p.podajWielkoscPuszki() << " mililitrow w cenie " << p.podajCene() << " zl" << endl;
		}
		// Metoda wywoywana podczas wizytacji obiektu klasy Wielopak
		virtual void wizytuj(Wielopak& w){
			cout << "+ Wielopak zawierajacy " << w.podajIlosc() << " nastepujacych artykulow: " << endl;
		}
};

/* Klasa demonstrujca uycia wizytatora jako obiektu kumulujcego stan obiektw wizytowanych. W tym przypadku
 * zliczana jest sumaryczna cena wszystkich odwiedzanych obiektw poprzez wywoanie metody 'podajCene', zdefiniowanej
 * w kadej z wizytowanych klas.
 */
class Zliczenie : public Wizytator{
	public:
		Zliczenie(){ suma = 0; }
		virtual void wizytuj(Chleb& c){
			suma += c.podajCene();
		}
		virtual void wizytuj(Piwo& p){
			suma += p.podajCene();
		}
		virtual void wizytuj(Wielopak& w){ // Sam wielopak nic nie kosztuje - cena rozoona jest na produkty
		}
		double podajSume(){
			return suma;
		}
	private:
		double suma;
};

int main(){
	vector<Produkt*> zakupy;
	Chleb c1(2.5,500);
	Piwo p1(2.9,600);
	Wielopak w1(10,4);
	// Utworzenie obiektu klasy vector reprezentujcego list zakupw.
	zakupy.push_back(&c1);
	zakupy.push_back(&p1);
	zakupy.push_back(&w1);
	// Obiekty wizytatorw zliczania wartoci i wypisywania
	Zliczenie zl;
	Wypisanie wp;
	// Wypisanie listy zakupw przy uyciu wizytatora wypisywania
	cout << "Lista zakupow: " << endl;
	for(vector<Produkt*>::iterator i = zakupy.begin(); i != zakupy.end(); i++)
		(*i)->akceptuj(wp);
	// Zliczenie sumarycznej wartoci produktw w koszyku przy uyciu wizytatora zlicznia
	for(vector<Produkt*>::iterator i = zakupy.begin(); i != zakupy.end(); i++)
		(*i)->akceptuj(zl);
	cout << "Sumaryczna wartosc produktow to: " << zl.podajSume() << " zl" << endl;
	return 0;
}

/* Wzorca wizytator wygodnie uywa si, gdy struktora obiektw wizytowanych jest staa, mog natomiast
 * zmienia si interfejsy poszczeglnych jej klas, np. poprzez dodanie nowej metody.
 * Jeli jednak to interfejs jest stay, natomiast zmianie podlego sama struktura klas, wygoniej jest
 * zdefiniowa odpowienie wywoania w samych klasach ni zmienia ca hierarchi klas wizytatorw
 */

